
// Application CommandLoop

package sptossg ;

import java.util.*;

public class CommandLoop {

    // MAIN - a routine for testing CompanyInfo storage and retrieval
    public static void main (String args[]) {
	new CommandLoop();
    }

    // I/O routines for the command loop
    String prompt(String pstr) {
	return prompt(pstr, null);
    }

    // I/O routines for the command loop

    //   Get a value from the user, using the passed in POSSVAL if there is
    //   one, but if not, then prompt the user for a value instead; A
    //   POSSVAL of "*" indicates a null value without prompting.
    String prompt(String pstr, String possval) {
	if (possval != null) {
	    possval = possval.trim();
	    if (possval.equals("*")) { return ""; }
	    if (! possval.equals("")) { return possval; }
	}

	byte linebytes[] = new byte[1024];
	if (pstr != null) {
	    System.out.print(pstr + ": ");
	    System.out.flush();
	}
	try {
	    System.in.read(linebytes);
	} catch (Exception exc) {
	    System.out.println("exception on read ");

	    return null;
	}

	String line = new String(linebytes);
	line = line.trim();
	if (line.equals("*")) { return ""; }
	return line;
    }


    // The following two functions, wordsep & getWord are used to parse items
    //   The idiom to pull a word off of a line is:
    // 		int pos = wordsep(line);
    // 		String word = getWord(line, pos);
    // 		line = line.substring(pos).trim();

    // separarate words by whitespace
    int wordsep(String line) {
	int last = line.length();

	if (last == 0) { return last; }

        // if the next word starts with '[', then ignore the whitespace and
        // scan till the ']' (but no nesting allowed)
	if (line.trim().charAt(0) == '[') {
	    int close = line.indexOf(']');
	    if (close >= 0 && close <= last) {
		last = close + 1;
	    }
	} else {
	    line = line + " \t";
	    last = Math.min(last,line.indexOf(' '));
	    last = Math.min(last,line.indexOf('\t'));
	}

	return last;
    }

    String getWord(String line, int sep) {
	String itm = line.substring(0,sep);
	itm = itm.trim();

	if (itm.equals("")) { return null; }
	if (itm.charAt(0) == '[') {
	    itm += "]";
	    int end = itm.indexOf(']');
	    itm = itm.substring(1, end);
	}

	return itm;
    }


    // cmdloop - A simple command processor for testing
    public CommandLoop() {
	CompanyInfo curcompany = null;
	Hashtable allcompanies = new Hashtable();

	System.out.println("\nSSG Program Command Line processor");
	System.out.println("Type 'help' for a list of commands\n");

	boolean prompt = true;

	while (true) {
	    String line;

	    line = prompt(prompt ? "Command" : null);

	    // A bracketed attribute is a shortcut for "dump [attr]"
	    if (line.startsWith("[")) { line = "dump " + line; }

	    int sep = wordsep(line);

	    String cmd = line.substring(0,sep).toLowerCase();
	    String params = line.substring(sep).trim();

	    // Exit the command loop
	    if (cmd.equals("quit")) { return; }

	    // list commands
	    if (cmd.equals("help")) {
		System.out.println("Available commands: ");
		System.out.println("File I/O routines");
		System.out.println("  parse - Parse an HTML file");
		System.out.println("  read - read a .SSG file");
		System.out.println("  write - write a .SSG file");
		System.out.println("  load - load internal definitions " +
				   "and dynamic datasets.");
		System.out.println("  save - save internal definitions " +
				   "and dynamic datasets.");
		System.out.println("Internal company manipulation");
		System.out.println("  list - List available companies");
		System.out.println("  rename - Rename a company");
		System.out.println("  use - Switch the current company");
		System.out.println("Data definition and examination");
		System.out.println("  def[ine] - Define a DynamicDataSet");
		System.out.println("  retr[ieve] - retrieve an attribute");
		System.out.println("  set - set an attribute");
		System.out.println("  attr - list the attributes");
		System.out.println("  dump - Display one or all attributes");
		System.out.println("  eval - evaluate a formula expression.");
		System.out.println("Other Commands");
		System.out.println("  echo - echo a line (for scripting)");
		System.out.println("  prompt - turn prompt on/off (for scripting)");
		System.out.println("  quit - exit the program");
	    }

	    // echo a line
	    if (cmd.equals("echo")) {
		System.out.println(params);
	    }

	    if (cmd.equals("prompt")) {
		prompt = ! prompt;
		System.out.println("Prompting is " + (prompt ? "on" : "off"));
	    }

	    // Define a DynamicDataSet
	    if (cmd.equals("def") || cmd.equals("define")) {

     		int pos = wordsep(params);
     		String word = getWord(params, pos);
     		params = params.substring(pos).trim();

 		String attr = prompt("Attribute Name", word);
 		String expr = prompt("Attribute Expression", params);

		DynamicDataSet dds = DynamicDataSet.define(attr, expr);
		if (dds == null) {
		    System.out.println("Parse error defining " + attr);
		} else {
		    System.out.println("Defined dynamic attribute: " + dds);
		}
	    }

	    // Initialize some formulas
	    if (cmd.equals("setup")) {

		String attrs[][] =
		{ {"PTP Margin", "100 * [PreTax Inc] / Sales"},
		  {"Sales Increase", "100 * Sales / prev(Sales)"}
		};

		for (int i = 0; i < attrs.length; i++) {
		    String aname = attrs[i][0];
		    String aformula = attrs[i][1];

		    DynamicDataSet.define(aname, aformula);
		    System.out.println("Defined dynamic attribute: " +
				       aname + " => " + aformula);
		}
	    }

	    // toggle debugging
	    if (cmd.equals("debug")) {
		Debug.toggle();
	    }

	    // Parse an HTML file
	    if (cmd.equals("parse")) {
		String coname = prompt("Company", params);

		curcompany = new CompanyInfo();
		allcompanies.put(coname, curcompany);

		try {
		    SPReportParser parser = new SPReportParser(coname + ".html");
		    if (!parser.parse(curcompany)) {
			System.err.println("parse() failed with the message: "
					   + parser.parseError());
		    } else {
			System.out.println(coname + ".html parsed. ");
		    }
		} catch (Exception exc) {
		    System.out.println("Error: " + exc);
		}
	    }

	    // read data from a .SSG file
	    if (cmd.equals("read")) {
		String coname = prompt("Company", params);

		curcompany = new CompanyInfo();
		allcompanies.put(coname, curcompany);

		curcompany.loadSSGfile(coname + ".ssg");
		System.out.println(coname + ".ssg read. ");
	    }

	    // write the current company out as a .SSG file
	    if (cmd.equals("write")) {
		if (curcompany == null) {
		    System.out.println("No current company.");
		} else {

		String ticker = curcompany.get("Symbol");
		String defstr = "";
		if (ticker != null) {
		    defstr = " (default is " + ticker + ")";
		}

		String coname = prompt("Company" + defstr, params);

		String fname = coname + ".ssg";

		curcompany.writeSSGfile(fname);
		System.out.println(fname + " written. ");
		}
	    }

	    // save global settings and dynamic attributes
	    if (cmd.equals("save")) {
		System.out.println("Can't save definitions yet. ");
		// TBD
	    }

	    // load global settings and dynamic attributes
	    if (cmd.equals("load")) {
		System.out.println("Can't load definitions yet. ");
		// TBD
	    }


	    // List available companies
	    if (cmd.equals("list")) {
		Enumeration e;

		int i = 0;
		e = allcompanies.keys();
		while (e.hasMoreElements()) {
		    String key = (String) e.nextElement();
		    System.out.println(i++ + ": " + key);
		}
		if (i == 0) {
		    System.out.println("No companies loaded.");
		}
	    }

	    // Rename a company
	    if (cmd.equals("rename")) {
		sep = wordsep(params);

		String coname = prompt("Company", params.substring(0,sep));
		String newname = prompt("New company id", params.substring(sep));

		CompanyInfo co = (CompanyInfo) allcompanies.get(coname);
		if (co == null) {
		    System.out.println("No company named '" + coname + "'");
		} else {
		    allcompanies.put(newname, co);
		    allcompanies.remove(coname);
		}
	    }

	    // Switch the current company
	    if (cmd.equals("use")) {
		String coname = prompt("Company", params);

		curcompany = (CompanyInfo) allcompanies.get(coname);
		if (curcompany == null) {
		    System.out.println("No company named '" + coname + "'");
		}
	    }

	    // retrieve or set an attribute in the current company
	    if (cmd.equals("retr") || cmd.equals("retrieve") ||
		cmd.equals("set")) {

		if (curcompany == null) {
		    System.out.println("No current company.");
		} else {

		    int pos = wordsep(params);
		    String word = getWord(params, pos);
		    params = params.substring(pos).trim();

		    String attr = prompt("Attribute", word);
		    String val;
		    String perref;

		    Period per;

		    pos = wordsep(params);
		    word = getWord(params, pos);
		    params = params.substring(pos).trim();

		    String year = prompt("Year", word);
		    if (year.equals("")) {
			per = new Period();
			perref = "";
		    } else {
			pos = wordsep(params);
			word = getWord(params, pos);
			params = params.substring(pos).trim();

			String period = prompt("Period", word);
			if (period.equals("")) {
			    per = new Period(year);
			    perref = "(" + year + ")";
			} else {
			    per = new Period(year, period);
			    perref = "(" + year + "," + period + ")";
			}
		    }

		    if (cmd.equals("set")) {
			val = prompt("Value", params);
			curcompany.set(val, attr, per);
		    } else {
			val = curcompany.get(attr, per);
		    }
		    System.out.println(attr + perref + " = " + val);
		}
	    }

	    // evaluate a formula in the current company
	    if (cmd.equals("eval") || cmd.equals("evaluate")) {

		if (curcompany == null) {
		    System.out.println("No current company.");
		} else {
		    double val;
		    String perref;

		    Period per;

		    int pos = wordsep(params);
		    String word = getWord(params, pos);
		    params = params.substring(pos).trim();

		    String year = prompt("Year", word);
		    if (year.equals("")) {
			per = new Period();
			perref = "";
		    } else {
			pos = wordsep(params);
			word = getWord(params, pos);
			params = params.substring(pos).trim();

			String period = prompt("Period", word);
			if (period.equals("")) {
			    per = new Period(year);
			    perref = "(" + year + ")";
			} else {
			    per = new Period(year, period);
			    perref = "(" + year + "," + period + ")";
			}
		    }

		    String expr = prompt("Expression", params);

		    try {
			val = Formula.eval(expr, curcompany, per);
			System.out.println("Expression" + perref + " = " + val);
		    } catch (Exception exc) {
			System.out.println("Parse Error: " + exc + "\nOn expression " + expr);
		    }
		}
	    }

	    // list the attributes of the current company
	    if (cmd.equals("attr")) {
		if (curcompany == null) {
		    System.out.println("No current company.");
		} else {
		    curcompany.dumpAttrs();
		}
	    }

	    // Display the current company
	    if (cmd.equals("dump")) {
		if (curcompany == null) {
		    System.out.println("No current company.");
		} else {
		    int pos = wordsep(params);
		    String word = getWord(params, pos);
		    params = params.substring(pos).trim();
		    String attr = prompt("Attribute", word);
		    if (attr.equals("")) {
			try {
			    curcompany.dump();
			} catch (Exception exc) {
			    System.out.println("Error dumping all attributes: " + exc);
			}
		    } else {
			try {
			    curcompany.dump(attr);
			} catch (Exception exc) {
			    System.out.println("Error dumping attribute " + attr + ": " + exc);
			}
		    }
		}
	    }
	}
    }

}
